from __future__ import annotations

__version__ = "0.1.2"

from pathlib import Path
from .chunker import Chunker
from collections.abc import Iterable, Callable
from typing import Any
import logging

logger = logging.getLogger(__name__)

def stream_polars_csv_gz(
    file_path: str | Path
    , buffer_size: int = 10_000_000
    , new_line_symbol: str = "\n"
    , func: Callable | None = None
    , schema: Any | None = None
    , verbose: bool = False
    , **kwargs
) -> Iterable[Any]:
    """
    Helper function that reads .csv.gz files in chunks. This requires Polars
    to be installed and should be version >= 1. This doesn't check Polars version
    for the user. The data schema, if not provided, will be inferred on the first chunk.
    The output type of this is an iterable of the output of `func`. If `func` 
    is None, then output will be an Iterable[pl.DataFrame]. If `func` returns (int, pl.DataFrame),
    then output will be Iterable[(int, pl.DataFrame)], etc.

    Please make sure that your system has > 10mb of RAM.

    Parameters
    ----------
    file_path
        Local file path or a temp file's name
    buffer_size
        Buffer size for the underlying chunker
    new_line_symbol
        The new line symbol for the .csv.gz file
    func
        An optional processor function that processes each chunk. The function signature
        should be func(df: pl.LazyFrame) -> Any. Notice the function input should
        be a lazy frame, because we can maximally optimize our operation on the chunk when it 
        is only scanned, not read into memory. Also note that if func is provided, 
        the output in the iterator are chunks we get after applying func to the original chunks.
    schema
        Schema of the dataset, if known. If none, this will be inferred on the first chunk. This must
        be a Polars-compatible Schema format.
    verbose
        Whether to print logs
    **kwargs
        Kwargs passed to Polars's scan_csv. Kwargs should not contain `has_header`, 
        and `schema`, since these are used internally.
    """
    import polars as pl

    if 'has_header' in kwargs:
        raise ValueError("Input `has_header` should not be a kwarg.")

    ck = Chunker(buffer_size=buffer_size, new_line_symbol=new_line_symbol).with_local_file(file_path, verbose=verbose)

    if schema is None:
        df_temp = pl.scan_csv(ck.read_one(), **kwargs) # first chunk
        use_schema = df_temp.collect_schema()
    else:
        df_temp = pl.scan_csv(ck.read_one(), schema=schema, **kwargs)
        use_schema = schema

    if func is None:
        yield df_temp.collect()
    else:
        yield func(df_temp)

    if verbose:
        logger.info(f"Number of reads: {ck.n_reads()}. Decompressed: {ck.bytes_decompressed()} bytes.")

    for byte_chunk in ck.chunks():
        if func is None:
            yield pl.read_csv(byte_chunk, has_header=False, schema=use_schema, **kwargs)
        else:
            yield func(
                pl.scan_csv(byte_chunk, has_header=False, schema=use_schema, **kwargs)
            )

        if verbose:
            logger.info(f"Number of reads: {ck.n_reads()}. Decompressed: {ck.bytes_decompressed()} bytes.")






from __future__ import annotations

from pathlib import Path
from enum import IntEnum

class CompressionMethod(IntEnum):
    UNK = 0 # Unknown or no compression 
    GZ = 1 # GZ


def _get_compression_method(b: bytes | str) -> CompressionMethod:

    if isinstance(b, str):
        _bytes = b.encode() # UTF-8
    else:
        _bytes = b

    if len(_bytes) < 2:
        return CompressionMethod.UNK
    else:
        if _bytes[:2] == b'\x1f\x8b':
            return CompressionMethod.GZ
        else:
            return CompressionMethod.UNK

def get_compression_method_local(path: str | Path) -> CompressionMethod:
    return _get_compression_method(
        open(path, "rb").read(2) # first two bytes
    )
    




from __future__ import annotations

import os
import sys
import logging

if sys.version_info >= (3, 11):
    from typing import Self
else:  # 3.10, 3.9, 3.8
    from typing_extensions import Self

from pathlib import Path
from typing import Iterator
from .utils import (
    CompressionMethod,
    get_compression_method_local
)
from gzeus._gzeus import (
    PyGzChunker,
)

__all__ = ["Chunker"]
logger = logging.getLogger(__name__)

class Chunker:
    
    def __init__(self, buffer_size:int = 1_000_000, new_line_symbol:str = "\n"):
        """
        Creates a Chunker.

        Parameters
        ----------
        buffer_size
            Buffer size in bytes. 1_000_000 is 1MB. A min of 1MB is used for this value.
            Note: in fact, the actual buffer that will get allocated has size slightly
            greater than the given value. Actual chunk size will vary depending on the compression, 
            but this buffer will be used repeatedly in the process, saving overall memory.
        new_line_symbol
            The symbol for new line. The last such symbol indicates the end of the 
            chunk. And the rest of the bytes will be appended to the front of the next
            chunk of bytes read.
        """    

        self.buffer_size = max(1_000_000, buffer_size)
        if len(new_line_symbol.encode('utf-8')) != 1:
            raise ValueError("The new line symbol must be one byte only.")

        self.symbol:str = new_line_symbol
        self.compression:CompressionMethod = CompressionMethod.UNK
        # Type of _reader is Option[ChunkReader], where ChunkReader represents types implementing the following interface
        # 1. is_finished(self) -> bool
        # 2. read_chunk(self) -> bytes
        # 3. read_full(self) -> bytes
        # 4. n_reads(self) -> usize
        # 5. bytes_decompressed(self) -> int
        # We use duck typing to make it work.
        self._reader = None
        self._description = ""

    def __repr__(self) -> str:
        desc = "Gz File Chunker:\n"
        if self._reader is None:
            desc += "Target file not set."
        else:
            desc += (
                f"Allocated internal buffer size: {self.buffer_size} bytes\n"
                f"New line symbol: {repr(self.symbol)}\n"
                f"{self._description}\n"
                "Read Status:\n"
                f"- # reads: {self._reader.n_reads()}\n"
                f"- Bytes decompressed: {self._reader.bytes_decompressed()}\n"
                f"- Job is finished: {self._reader.is_finished()}"
            )

        return desc

    def _check_reader(self):
        """
        Checks whether the internal reader is set.
        """
        # Always run this when a read is called
        if self._reader is None:
            raise ValueError("Target file is not set yet. Please run `with_*_file` first.")

        if self._reader.is_finished():
            raise ValueError("The reader has finished reading. To begin a new read, please run ``with_*_file` again.")

    def with_buffer_size(self, buffer_size:int) -> Self:
        """
        Resets the buffer size of the Chunker.

        Parameters
        ----------
        buffer_size
            The internal buffer size. It should be at least 1_000_000, which is 1MB. Note that
            if this is too high, your network might complain and you may have have disconnect issues
            when reading from cloud storage like s3.
        """
        if buffer_size <= 1_000_000:
            self.buffer_size = 1_000_000
        else:
            self.buffer_size = buffer_size

        return self

    def with_new_line(self, new_line_symbol:str) -> Self:
        """
        Resets the new line symbol.

        Parameters
        ----------
        new_line_symbol
            The new line symbol for the underlying text file. The most common one is '\n'.
        """

        if len(new_line_symbol) != 1:
            raise ValueError("The new line symbol must be one byte only.")

        self.symbol = new_line_symbol
        return self

    def with_local_file(self, file_path: str | Path, verbose:bool=False) -> Self:
        """
        Prepares the chunker by letting it know the file to be read.

        Parameters
        ----------
        file_path
            The file path
        verbose
            Whether to print more info about the job
        """
        self.compression = get_compression_method_local(file_path)
        if self.compression == CompressionMethod.GZ:
            self._reader = PyGzChunker(str(file_path), self.buffer_size, self.symbol)
        else:
            raise ValueError(
                "The underlying file is not compressed or cannot be properly identified as a .gz file."
            )

        self._description = f"Target file is a local file at path: {file_path}"
        if verbose:
            with open(file_path, "rb") as f:
                # place the cursor in the right position
                f.seek(-4, os.SEEK_END)
                # get the size of uncompressed input from last 4 bytes
                size = int.from_bytes(f.read(), "little" )
                logger.info(f"Estimated uncompressed size is: {size} bytes, estimated # chunks needed: {int(size/self.buffer_size)}. (This estimate can be wrong if the file is a Multi-file gz instead of a single-file gz.)")

        return self

    def n_reads(self) -> int:
        """
        Return the number of chunks read.
        """
        if self._reader is None:
            return 0
        return self._reader.n_reads()

    def is_finished(self) -> bool:
        """
        Whether the current reader has finished reading.
        """
        if self._reader is None:
            return False
        return self._reader.is_finished()

    def bytes_decompressed(self) -> int:
        """
        Returns the number of decompressed bytes so far by the reader.
        """
        if self._reader is None:
            return 0
        return self._reader.bytes_decompressed()

    def read_full(self) -> bytes:
        """
        Read as much as the internal buffer allows, and after this read, declare the read to be finished. 
        This will read at most self.buffer_size number of bytes, and should only be used when you know that 
        the decompressed file is small enough to fit in the buffer. This should be used for convenience only.
        """
        _ = self._check_reader()
        return self._reader.read_full()

    def read_one(self) -> bytes:
        """
        Read one chunk.
        """
        _ = self._check_reader()
        return self._reader.read_chunk()

    def chunks(self) -> Iterator[bytes]:

        _ = self._check_reader()
        while not self._reader.is_finished():
            bytes_read = self._reader.read_chunk()
            if len(bytes_read) > 0:
                yield bytes_read